/*
 * $Header: /cvsroot/jroseviewer/jroseviewer/src/rv/parser/petal.g,v 1.14 2003/01/23 06:32:09 jackh Exp $
 * $Revision: 1.14 $
 * $Date: 2003/01/23 06:32:09 $
 */
header {
package rv.parser;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.Vector;

import org.apache.log4j.Logger;

import rv.petal.*;
}

/**
 * A parser for parsing petal files.<br>
 * <b>This class is generated by ANTLR.  DO NOT MODIFIY THIS CLASS MANUALLY.</b>
 *
 * @author Jack Hong
 * @version $Id: petal.g,v 1.14 2003/01/23 06:32:09 jackh Exp $
 */
class PetalParser extends Parser;
options {
	exportVocab=Petal;
	k = 3;
}

{
	/** An instance of <code>Logger</code>. */
	private static Logger log = Logger.getLogger(PetalParser.class);
	
	/** The path-separator used in petal files. */
	public final static String PATH_SEPARATOR = "\\\\";
	
	/** The file to be parsed. */
	private File currentFile = null;
	
	/**
	 * An instance of <code>PathMap</code> used to resolve path symbols.
	 */
	private PathMap pm = null;
	
	/** A map that contains all elements. */
	private HashMap uniqueIdsMap = new HashMap();
	
	/**
	 * Returns the <code>PathMap</code> used in the parser. 
	 */
	public PathMap getPathMap() {
		return pm;
	}
	
	/**
	 * Sets the <code>PathMap</code> used in the parser.
	 */
	public void setPathMap(PathMap pm) {
		this.pm = pm;
	}
	
	/** Returns the uniqueIdsMap. */
	public final HashMap getUniqueIdsMap() {
		return uniqueIdsMap;
	}

	/**
	 * Parses the specified <i>file</i>.
	 * @param file a file to be parsed.
	 * @return an instance of <code>PetalFile</code>.
	 */
	public static PetalFile parse(File file, PathMap pm) throws PetalParserException {
		try {
			FileInputStream fis = new FileInputStream(file);

			PetalLexer lexer = new PetalLexer(fis);

			PetalParser parser = new PetalParser(lexer);
			parser.setPathMap(pm);
			parser.currentFile = file;

			PetalFile petalFile = parser.petal_file();

			fis.close();
			
			return petalFile;
		}
		catch (FileNotFoundException fnfe) {
			log.fatal(fnfe);
			throw new PetalParserException("Could not found " + file.getPath());
		}
		catch (IOException ie) {
			log.fatal(ie);
			throw new PetalParserException("IOException occurred on " + file.getPath());
		}
		catch (RecognitionException re) {
			log.fatal(re);
			throw new PetalParserException("RecognitionException occurred on " + file.getPath());
		}
		catch (TokenStreamException te) {
			log.fatal(te);
			throw new PetalParserException("TokenStreamException occurred on " + file.getPath());
		}
	}
	
	/**
	 * Parses the specified <i>file</i>.
	 */
	public static PetalFile parse(String filename, PathMap pm) throws PetalParserException {
		return parse(new File(filename), pm);
	}
	
	/**
	 * Parses an external file.
	 */
	protected PetalFile resolveExternalFile(String extfile) {
		String ref = convertFilePath(extfile);
		String filepath = null;
		
		if (ref.startsWith("$")) {
			String symbol = null;
			String actual = null;
			String name = null;
			
			int i = ref.indexOf(File.separator);
			
			if (i != -1) {
				symbol = ref.substring(1, i);
				name = ref.substring(i + 1);

				if ("CURDIR".equals(symbol)) {
					String currentPath = currentFile.getPath();
					int n = currentPath.lastIndexOf(File.separator);
					actual = currentPath.substring(0, n + 1);
				}
				else if (pm instanceof PathMap) {
					actual = pm.getActualPath(symbol);

					if (actual instanceof String) {
						if (!actual.endsWith(File.separator))
							actual = actual + File.separator;
					}
					else {
						log.error("could not find the actual path of symbol: $" + symbol);
						return null;
					}
				}
				else {
					log.error("unknown path symbol: $" + symbol);
					return null;
				}
				
				filepath = actual + name;
			}
			else {
				log.error("invalid path symbol: " + ref);
				return null;
			}
		}
		else {
			filepath = ref;
		}
		
		try {
			log.debug("parsing external file: " + filepath);
			PetalFile pf = parse(filepath, getPathMap());
			log.debug("done.");
			return pf;
		}
		catch (PetalParserException pe) {
			log.error("Could not parse external file: " + filepath);
			return null;
		}
	}
	
	/**
	 * Converts the specified <i>path</i> to be recognizable by the host 
	 * operating system.
	 */
	protected String convertFilePath(String path) {
		int slen = PATH_SEPARATOR.length();

		StringBuffer buf = new StringBuffer();
		int s = 0;
		int i = 0;
		
		while (true) {
			i = path.indexOf(PATH_SEPARATOR, s);
			if (i != -1) {
				buf.append(path.substring(s, i));
				buf.append(File.separator);
				s = i + slen;
			}
			else {
				buf.append(path.substring(s));
				break;
			}
		}
		
		return buf.toString();
	}
}


petal_file returns [PetalFile pfile]
	{
		pfile = new PetalFile();

		Element petal = null;
		Element root = null;
	}
	:	petal=object[pfile] { pfile.setPetalFileInfo((PetalFileInfo) petal); }
		root=object[pfile] { pfile.setRootElement(root); }
		EOF!
		{
			pfile.setUniqueIdsMap(uniqueIdsMap);

			if (root instanceof Model) {
				Model model = (Model) root;
				model.setUniqueIdsMap(uniqueIdsMap);
			}
		}
	;

object [RoseObject parent] returns [Element element]
	{
		HashMap properties = new HashMap();
		element = null;
		
		PetalObjectParameters params = null;
		Integer tagValue = null;
		PetalAttribute attr = null;

	}
	:	OBJECT type:IDENT (params=object_parameters)? (tagValue=tag)?
		{
			String objectType = type.getText();
			element = PetalFactory.createElement(objectType);
			element.setParent(parent);
		}
			(attr=attribute[element]
				{
					Object value = attr.getValue();
					if (value instanceof PetalLocation) {
						if (element instanceof RoseItemView) {
							RoseItemView view = (RoseItemView) element;
							PetalLocation location = (PetalLocation) value;
							view.setLocation(location.getX(), location.getY());
							
							if (tagValue instanceof Integer)
								view.setTag(tagValue.intValue());
						}
						else {
							log.warn("unknown location property for " + element + " discarded.");
						}
					}
					
					properties.put(attr.getAttribute(), attr.getValue()); }
			)*
		RPAREN
		{
			element.setProperties(properties);

			if (params instanceof PetalObjectParameters)
				element.setName(params.getName());

			String uniqueId = element.getUniqueId();
			if (uniqueId instanceof String && uniqueId.length() > 0) {
				uniqueIdsMap.put(uniqueId, element);
			}

			if (element instanceof ControllableUnit)
			{
				ControllableUnit unit = (ControllableUnit) element;
				if (!unit.isLoaded()) {
					PetalFile pf = resolveExternalFile(unit.getFileName());
					if (pf instanceof PetalFile) {
						uniqueIdsMap.putAll(pf.getUniqueIdsMap());
						Element rootElement = pf.getRootElement();
						rootElement.setParent(parent);
						return rootElement;
					}
				}
			}
			
		}
	;

object_parameters returns [PetalObjectParameters params]
	{
		params = new PetalObjectParameters();
	}
	:	(name:STRING 
			{
				params.setName(name.getText());
			} 
		)
	|	(category:STRING name1:STRING 
			{
				params.setCategory(category.getText());
				params.setName(name1.getText());
			}
		)
	|	(name2:STRING p1:STRING p2:STRING 
			{
				params.setName(name2.getText());
				params.setParameter1(p1.getText());
				params.setParameter2(p2.getText());
			}
		)
	;

tag returns [Integer t]
	{
		t = null;
	}
	:	TAG! i:INT
		{ t = new Integer(i.getText()); }
	;

attribute [Element parent] returns [PetalAttribute attr]
	{
		attr = new PetalAttribute();

		Object r = null;
		Element e = null;
		PetalList l = null;
		PetalValue v = null;
		PetalLocation c = null;
		Integer t = null;
		PetalValue p = null;
	}
	:	n:IDENT
			{ attr.setAttribute(new String(n.getText())); }
		(r=literal
			{ attr.setValue(r); }
		| e=object[parent] 
			{ attr.setValue(e); }
		| l=list[parent] 
			{ attr.setValue(l.getElements()); }
		| v=value 
			{ attr.setValue(v); }
		| c=location 
			{ attr.setValue(c); }
		| t=tag 
			{ attr.setValue(t); }
		| p=value_pair
			{ attr.setValue(p); }
		)
	;

location returns [PetalLocation l]
	{
		l = null;
	}
	:	LPAREN x:INT COMMA y:INT RPAREN
		{
			l = new PetalLocation(Integer.parseInt(x.getText()), 
				Integer.parseInt(y.getText()));
		}
	;

list [Element parent] returns [PetalList list]
	{
		list = null;
		Vector elements = new Vector();
		Element e = null;
		PetalLocation l = null;
	}
	:	LIST (n:IDENT)?
		(e=object[parent]
			{ elements.add(e); }
		| l=location
			{ elements.add(l); }
		)*
		RPAREN
		{
			String id = null;
			if (n != null)
				id = n.getText();
			
			list = new PetalList(id, elements);
		}
	;

value returns [PetalValue v]
	{
		v = new PetalValue();
		Object l = null;
	}
	:	VALUE key:IDENT l=literal
		RPAREN
		{ v = new PetalValue(key.getText(), l); }
	;

value_pair returns [PetalValue v]
	{
		v = null;
	}
	:	LPAREN key:STRING i:INT RPAREN
		{ v = new PetalValue(key.getText(), new Integer(i.getText())); }
	;

literal returns [Object l]
	{
		l = null;

		Integer i = null;
		Float f = null;
		Boolean b = null;
		String s = null;
		String m = null;
	}
	:	i=integer_literal {l = i;}
	|	f=float_literal {l = f;}
	|	b=boolean_literal {l = b;}
	|	s=string_literal {l = s;}
	|	m=ml_string_literal {l = m;}
	;

integer_literal returns [Integer i]
	{
		i = null;
	}
	:	n:INT
		{ i = new Integer(n.getText()); }
	;

float_literal returns [Float f]
	{
		f = null;
	}
	:	n:FLOAT
		{ f = new Float(n.getText()); }
	;

boolean_literal returns [Boolean b]
	{
		b = null;
	}
	:	"TRUE"
		{ b = new Boolean(true); }
	|	"FALSE"
		{ b = new Boolean(false); }
	;

string_literal returns [String str]
	{
		str = null;
	}
	:	s:STRING
		{ str = new String(s.getText()); }
	;

ml_string_literal returns [String str]
	{
		str = null;

		StringBuffer buf = new StringBuffer();
	}
	:	(s:ML_STRING
			{
				buf.append(s.getText());
				buf.append("\n");
			}
		)+
		{ str = buf.toString(); }
	;



/**
 * A lexer for lexing petal files.<br>
 * <b>This class is generated by ANTLR.  DO NOT MODIFIY THIS CLASS MANUALLY.</b>
 *
 * @author Jack Hong
 * @version $Id: petal.g,v 1.14 2003/01/23 06:32:09 jackh Exp $
 */
class PetalLexer extends Lexer;
options {	
	exportVocab=Petal;
	charVocabulary = '\3'..'\377';
	testLiterals=false;
	k = 3;
}

// punctuation
COMMA    :	',';

LPAREN   :	'(';

RPAREN   :	')';

TAG      :	'@';

// entities
OBJECT
options {
	paraphrase = "an object";
}
	:	LPAREN "object"
	;

LIST
options {
	paraphrase = "a list";
}
	:	LPAREN "list"
	;

VALUE
options {
	paraphrase = "a value";
}
	:	LPAREN "value"
	;


// identifiers
IDENT
options {
	paraphrase = "an identifier";
	testLiterals=true;
}
	:	( 'a'..'z' | 'A'..'Z' | '_') ( 'a'..'z' | 'A'..'Z' | '_' | '0'..'9')*
	;

// whitespace -- ignored
WS	:	(	' '
		|	'\t'
		|	'\f'
		// handle newlines
		|	(	"\r\n"  // DOS
			|	'\r'    // Macintosh
			|	'\n'    // Unix
			)
			{ newline(); }
		)
		{ $setType(Token.SKIP); }
	;

// float literals
FLOAT
options {
	paraphrase = "a floating point value";
}
	:	'.' (DIGIT)+
	;

// integer literals
INT
options {
	paraphrase = "an integer value";
}
	:	('+' | '-')? (DIGIT)+ 
		('.' (DIGIT)* 
			{
				$setType(FLOAT);
			}
		)?
	;

protected
DIGIT
options {
	paraphrase = "a digit";
}
	:	'0'..'9'
	;

// string literals
STRING
options {
	paraphrase = "a string";
}
	:	'"'! (~'"')* '"'!
	;

// multi-line string literals
ML_STRING
options {
	paraphrase = "a multi-line string";
}
	:	'|'! (~('\r' | '\n'))*
	;

